/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2017, Natural Language Processing Lab, Northestern University. 
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: Xu Chen (email: hello_master1954@163.com) 2018-09-17
 */

#include <math.h>
#include "CrossEntropy.h"
#include "CrossEntropy.cuh"
#include "../core/arithmetic/MultiplyDim.h"
#include "../core/arithmetic/Multiply.h"
#include "../core/math/Unary.h"
#include "../core/math/ScaleAndShift.h"
#include "../core/arithmetic/Negate.h"
#include "../core/reduce/ReduceSum.h"
#include "../core/reduce/ReduceSumAll.h"

namespace nts{ // namespace nts(NiuTrans.Tensor)

/*
compute the cross entropy loss

loss = sum_{i} (-gold_i * log(output_i))
where gold and output are distributions 
        
>> output - model prediction
>> gold - gold standard
>> loss - compute loss
>> weight - a rescaling weight given to each class
>> padding - specify a target value that is ignored and does not contribute to the loss computation
>> leadingDim - the leading dimension for the output
*/
void _CrossEntropy(const XTensor * output, const XTensor * gold,
                   XTensor * loss, const XTensor * weight, 
                   const XTensor * padding, int leadingDim)
{
    int n = leadingDim < 0 ? output->order - 1 : leadingDim;
    CheckNTErrors(n >= 0 && n < output->order, "Wrong leadingDim!");

    int unitNum = output->dimSize[n];

    CheckNTErrors(XTensor::IsSameShaped(output, gold), 
                 "The output tensor and gold tensor must be of the same size!");
    CheckNTErrors(weight == NULL || weight->unitNum == unitNum, "Wrong weight tensor!");
    CheckNTErrors(padding == NULL || XTensor::IsSameShaped(padding, loss), "The loss tensor and padding tensor must be same shape!");
    CheckNTErrors(loss->order == output->order - 1, "Wrong loss dimension!");
    CheckNTErrors(gold->dataType == DEFAULT_DTYPE && output->dataType == DEFAULT_DTYPE, "TODO!");

    XTensor * logInter = NewTensorBuf(output, output->devID, output->mem);
    XTensor * mulInter = NewTensorBuf(output, output->devID, output->mem);
    XTensor * negInter = NewTensorBuf(output, output->devID, output->mem);

    XTensor * logBuf = NewTensorBuf(output, output->devID, output->mem);
    XTensor * mulBuf = NewTensorBuf(output, output->devID, output->mem);
    XTensor * negBuf = NewTensorBuf(output, output->devID, output->mem);

    /* l = log(output) */
    _Log(output, logBuf);

    if(weight != NULL){
        XTensor * weightBuf = NewTensorBuf(output, output->devID, output->mem);
        /* multiply gold and weight by broadcast wg = mulDim(g * w) */
        _MultiplyDim(gold, weight, weightBuf, n, 0);
        /* multiply weighted gold and log(output) wgl = mul(wg, l) */
        _Multiply(weightBuf, logBuf, mulBuf, 0);
        DelTensorBuf(weightBuf);
    }
    else{
        /* multiply gold and log(output) gl = mul(g, l) */
        _Multiply(gold, logBuf, mulBuf, 0);
    }

    /* negate multiply result n = negate(mul) */
    _NegateMe(mulBuf);
    
    _ReduceSum(mulBuf, loss, n);

    DelTensorBuf(negInter);
    DelTensorBuf(mulInter);
    DelTensorBuf(logInter);
}

/*
compute the cross entropy loss (implementation manually)

loss = sum_{i} (-gold_i * log(output_i))
where gold and output are distributions 
        
>> output - model prediction
>> gold - gold standard
>> loss - compute loss
>> weight - a rescaling weight given to each class
>> padding - specify a target value that is ignored and does not contribute to the loss computation
>> leadingDim - the leading dimension for the output
*/
void _CrossEntropyManual(const XTensor * output, const XTensor * gold,
                         XTensor * loss, const XTensor * weight, 
                         const XTensor * padding, int leadingDim)
{
#ifdef USE_CUDA
    if(output->devID >= 0) {
        _CudaCrossEntropyManual(output, gold, loss, weight, padding, leadingDim);
        return;
    }
#endif

    int order = output->order;
    int n = leadingDim < 0 ? output->order - 1 : leadingDim;
    int leadingDimSize = output->GetDim(n);

    CheckNTErrors(n >= 0 && n < output->order, 
                 "Wrong leadingDim!");
    CheckNTErrors(XTensor::IsSameShaped(output, gold), 
                 "The output tensor and gold tensor must be of the same size!");
    CheckNTErrors(weight == NULL || weight->unitNum == leadingDimSize, 
                 "Wrong weight tensor!");
    CheckNTErrors(padding == NULL || XTensor::IsSameShaped(padding, loss), 
                 "The loss tensor and padding tensor must be same shape!");
    CheckNTErrors(loss->order == output->order - 1, 
                 "Wrong loss dimension!");
    CheckNTErrors(gold->dataType == DEFAULT_DTYPE && output->dataType == DEFAULT_DTYPE, 
                 "TODO!");
    
    int blockNum = 1;
    int blockSize = 1;
    int stride = 1;

    for(int i = n + 1; i < order; i++)
        stride *= output->GetDim(i);
    
    blockSize = stride * leadingDimSize;
    blockNum = output->unitNum / blockSize;

    DTYPE * outputData = (DTYPE*)output->data;
    DTYPE * goldData = (DTYPE*)gold->data;
    DTYPE * lossData = (DTYPE*)loss->data;

    DTYPE tmpLoss;
    if(weight == NULL) {
        if(padding == NULL) {
            for(int i = 0; i < blockNum; i++) {
                int beg = i * blockSize;

                tmpLoss = 0;
                for(int j = 0; j < blockSize; j++) 
                    tmpLoss += -(*(goldData + beg + j)) * 
                                (DTYPE)log(*(outputData + beg + j));
                *(lossData + i) = tmpLoss;
            }
        }
        else {
            DTYPE * paddingData = (DTYPE*)padding->data;
            for(int i = 0; i < blockNum; i++) {
                int beg = i * blockSize;

                if(*(paddingData + i) == 0)
                    *(lossData + i) = 0;
                else{
                    tmpLoss = 0;
                    for(int j = 0; j < blockSize; j++)
                        tmpLoss += -(*(goldData + beg + j)) * 
                                    (DTYPE)log(*(outputData + beg + j));
                    *(lossData + i) = tmpLoss;
                }
            }            
        }
    }
    else {
        DTYPE * weightData = (DTYPE*)weight->data;
        if(padding == NULL) {
            for(int i = 0; i < blockNum; i++) {
                int beg = i * blockSize;

                tmpLoss = 0;
                for(int j = 0; j < blockSize; j++)
                    tmpLoss += -(*(goldData + beg + j)) * 
                                (DTYPE)log(*(outputData + beg + j)) * 
                                (*(weightData + j));
                *(lossData + i) = tmpLoss;
            }
        }
        else {
            DTYPE * paddingData = (DTYPE*)padding->data;
            for(int i = 0; i < blockNum; i++) {
                int beg = i * blockSize;

                if(*(paddingData + i) == 0)
                    *(lossData + i) = 0;
                else{
                    tmpLoss = 0;
                    for(int j = 0; j < blockSize; j++)
                        tmpLoss += -(*(goldData + beg + j)) * 
                                    (DTYPE)log(*(outputData + beg + j)) * 
                                    (*(weightData + j));
                    *(lossData + i) = tmpLoss;
                }
            }              
        }
    }
}

/*
get the dimSize after reduce operation
>> tensor - a tensor to be reduced
>> n - the reduce dimension 
<< return - the pointer of dimSize
*/
int * reduceDimSize(const XTensor * tensor, int n)
{
    int order = tensor->order;
    int * dimSize = new int[order - 1];

    for (int i = 0; i < order; i++) {
        if(i < n)
            dimSize[i] = tensor->dimSize[i];
        else if(i > n)
            dimSize[i - 1] = tensor->dimSize[i];
    }
    return dimSize;
}

/*
compute the cross entropy loss
loss = sum_{i} (-gold_i * log(output_i))
where gold and output are distributions 
        
>> output - model prediction
>> gold - gold standard
>> reduce - loss compute way, sum or mean
>> weight - a rescaling weight given to each class
>> padding - specify a target value that is ignored and does not contribute to the loss computation
>> leadingDim - the leading dimension for the output
*/
DTYPE _CrossEntropy(const XTensor * output, const XTensor * gold,
                    LOSS_COMPUTE_WAY reduceWay, const XTensor * weight, 
                    const XTensor * padding, int leadingDim)
{
    int n = leadingDim < 0 ? output->order - 1 : leadingDim;
    CheckNTErrors(n >= 0 && n < output->order, "Wrong leadingDim!");

    int unitNum = output->dimSize[n];

    CheckNTErrors(XTensor::IsSameShaped(output, gold), 
                 "The output tensor and gold tensor must be of the same size!");
    CheckNTErrors(weight == NULL || weight->unitNum == unitNum, "Wrong weight tensor!");
    CheckNTErrors(padding == NULL || padding->order == output->order - 1, "The loss tensor and padding tensor must be same shape!");
    CheckNTErrors(gold->dataType == DEFAULT_DTYPE && output->dataType == DEFAULT_DTYPE, "TODO!");

    XTensor * logBuf = NewTensorBuf(output, output->devID, output->mem);
    XTensor * mulBuf = NewTensorBuf(output, output->devID, output->mem);
    XTensor * negBuf = NewTensorBuf(output, output->devID, output->mem);

    /* l = log(output) */
    _Log(output, logBuf);

    if(weight != NULL){
        XTensor * weightBuf = NewTensorBuf(output, output->devID, output->mem);
        /* multiply gold and weight by broadcast wg = mulDim(g * w) */
        _MultiplyDim(gold, weight, weightBuf, n, 0);
        /* multiply weighted gold and log(output) wgl = mul(wg, l) */
        _Multiply(weightBuf, logBuf, mulBuf, 0);
        DelTensorBuf(weightBuf);
    }
    else{
        /* multiply gold and log(output) gl = mul(g, l) */
        _Multiply(gold, logBuf, mulBuf, 0);
    }

    /* negate multiply result n = negate(mul) */
    _NegateMe(mulBuf);
    
    int * dimSize;
    dimSize = reduceDimSize(output, n);
    XTensor * lossInter = NewTensor(output->order - 1, dimSize, output->dataType, output->denseRatio, output->devID, output->mem);
    
    /* reduce sum all classes */
    _ReduceSum(mulBuf, lossInter, n);

    DelTensorBuf(negBuf);
    DelTensorBuf(mulBuf);
    DelTensorBuf(logBuf);
    
    DTYPE loss;

    /* compute the total loss */
    if(padding != NULL) {
        XTensor * temp(lossInter);
        _Multiply(lossInter, padding, temp);
        loss = _ReduceSumAll(temp);
        delete temp;
    }
    else 
        loss = _ReduceSumAll(lossInter);

    if(reduceWay == REDUCE_MEAN) {
        if(padding != NULL) {
            XTensor * zeroIndicator = NewTensorBuf(padding, padding->devID, padding->mem);
            
            _IsZero(padding, zeroIndicator);
            int reduceSize = (int)_ReduceSumAll(zeroIndicator);
            loss = loss / (DTYPE)(padding->unitNum - reduceSize);

            DelTensorBuf(zeroIndicator);
        }
        else 
            loss = loss / (DTYPE)lossInter->unitNum;
    }
    else if(reduceWay == REDUCE_SUM) {
        /* don't need to do anything */
    }
    else {
        ShowNTErrors("TODO");
    }

    delete[] dimSize;
    delete lossInter;

    return loss;
}

/*
compute the cross entropy loss (implementation manually)

loss = sum_{i} (-gold_i * log(output_i))
where gold and output are distributions 
        
>> output - model prediction
>> gold - gold standard
>> reduceWay - loss compute way, sum or mean
>> weight - a rescaling weight given to each class
>> padding - specify a target value that is ignored and does not contribute to the loss computation
>> leadingDim - the leading dimension for the output
<< return - the cross entropy loss that is a scalar
*/
DTYPE _CrossEntropyManual(const XTensor * output, const XTensor * gold,
                          LOSS_COMPUTE_WAY reduceWay, const XTensor * weight, 
                          const XTensor * padding, int leadingDim)
{
#ifdef USE_CUDA
    if(output->devID >= 0) {
        return _CudaCrossEntropyManual(output, gold, reduceWay, weight, padding, leadingDim);
    }
#endif

    int order = output->order;
    int n = leadingDim < 0 ? output->order - 1 : leadingDim;
    int leadingDimSize = output->GetDim(n);

    CheckNTErrors(n >= 0 && n < output->order, 
                 "Wrong leadingDim!");
    CheckNTErrors(XTensor::IsSameShaped(output, gold), 
                 "The output tensor and gold tensor must be of the same size!");
    CheckNTErrors(weight == NULL || weight->unitNum == leadingDimSize, 
                 "Wrong weight tensor!");
    CheckNTErrors(padding == NULL || padding->order == output->order - 1, 
                 "Wrong padding tensor!");
    CheckNTErrors(gold->dataType == DEFAULT_DTYPE && output->dataType == DEFAULT_DTYPE, 
                 "TODO!");
    
    int blockNum = 1;
    int blockSize = 1;
    int stride = 1;

    for(int i = n + 1; i < order; i++)
        stride *= output->GetDim(i);
    
    blockSize = stride * leadingDimSize;
    blockNum = output->unitNum / blockSize;

    DTYPE * outputData = (DTYPE*)output->data;
    DTYPE * goldData = (DTYPE*)gold->data;

    DTYPE loss = 0;
    int nonZeroNum = 0;

    if(weight == NULL) {
        if(padding == NULL) {
            nonZeroNum = blockNum;
            for(int i = 0; i < blockNum; i++) {
                int beg = i * blockSize;

                for(int j = 0; j < blockSize; j++) 
                    loss += -(*(goldData + beg + j)) * 
                             (DTYPE)log(*(outputData + beg + j));
            }
        }
        else {
            DTYPE * paddingData = (DTYPE*)padding->data;
            for(int i = 0; i < blockNum; i++) {
                if(*(paddingData + i) == 0)
                    continue;
                else{
                    nonZeroNum += 1;

                    int beg = i * blockSize;
                    for(int j = 0; j < blockSize; j++)
                        loss += -(*(goldData + beg + j)) * 
                                 (DTYPE)log(*(outputData + beg + j));
                }
            }            
        }
    }
    else {
        DTYPE * weightData = (DTYPE*)weight->data;
        if(padding == NULL) {
            nonZeroNum = blockNum;
            for(int i = 0; i < blockNum; i++) {
                int beg = i * blockSize;
                for(int j = 0; j < blockSize; j++)
                    loss += -(*(goldData + beg + j)) * 
                             (DTYPE)log(*(outputData + beg + j)) * 
                             (*(weightData + j));
            }
        }
        else {
            DTYPE * paddingData = (DTYPE*)padding->data;
            for(int i = 0; i < blockNum; i++) {
                if(*(paddingData + i) == 0)
                    continue;
                else{
                    nonZeroNum += 1;

                    int beg = i * blockSize;
                    for(int j = 0; j < blockSize; j++)
                        loss += -(*(goldData + beg + j)) * 
                                 (DTYPE)log(*(outputData + beg + j)) * 
                                 (*(weightData + j));
                }
            }              
        }
    }

    if(reduceWay == REDUCE_MEAN) {
        loss = loss / (DTYPE)nonZeroNum;
    }
    else if(reduceWay == REDUCE_SUM) {
        /* don't need to do anything */
    }
    else {
        ShowNTErrors("TODO");
    }

    return loss;
}

/* 
backward compuation for cross entropy function (tensor version)

loss = sum_{i} (-t_i * log(y_i))
dE/dy_i = -t_i / y_i
where E is the error(loss) function that measure the errors in y
with respect to gold standard, and y this the model output

>> dedy - dE/dy (for return)
>> output - model prediction
>> gold - gold standard
>> weight - a rescaling weight given to each class
>> padding - specify a target value that is ignored and does not contribute to the loss computation
>> leadingDim - the leading dimension for the output
*/
void _CrossEntropyBackward(XTensor * dedy, const XTensor * output, const XTensor * gold, 
                           const XTensor * weight, XTensor * padding, 
                           int leadingDim)
{
#ifdef USE_CUDA
    if(output->devID >= 0) {
        _CudaCrossEntropyBackward(dedy, output, gold, weight, padding, leadingDim);
        return;
    }
#endif

    int order = output->order;
    int n = leadingDim < 0 ? output->order - 1 : leadingDim;
    int leadingDimSize = output->GetDim(n);
    int unitSize = dedy->unitSize;

    CheckNTErrors(n >= 0 && n < output->order, 
                 "Wrong leading dimension!");
    CheckNTErrors(XTensor::IsSameShaped(dedy, output, gold), 
                 "The output tensor and gold tensor must be of the same size!");
    CheckNTErrors(weight == NULL || weight->unitNum == leadingDimSize, 
                 "Wrong weight tensor!");
    CheckNTErrors(padding == NULL || padding->order == output->order - 1, 
                 "Wrong padding tensor!");
    CheckNTErrors(gold->dataType == DEFAULT_DTYPE && output->dataType == DEFAULT_DTYPE, 
                 "TODO!");
    
    int blockNum = 1;
    int blockSize = 1;
    int stride = 1;

    for(int i = n + 1; i < order; i++)
        stride *= output->GetDim(i);
    
    blockSize = stride * leadingDimSize;
    blockNum = output->unitNum / blockSize;

    DTYPE * dedyData = (DTYPE*)dedy->data;
    DTYPE * outputData = (DTYPE*)output->data;
    DTYPE * goldData = (DTYPE*)gold->data;

    if(weight == NULL) {
        if(padding == NULL) {
            for(int i = 0; i < blockNum; i++) {
                int beg = i * blockSize;
                for(int j = 0; j < blockSize; j++)
                    *(dedyData + beg + j) = -(*(goldData + beg + j)) / 
                                             (*(outputData + beg + j));
            }
        }
        else {
            DTYPE * paddingData = (DTYPE*)padding->data;
            for(int i = 0; i < blockNum; i++) {
                int beg = i * blockSize;
                if(*(paddingData + i) == 0)
                    memset(dedyData + beg, 0, blockSize * unitSize);
                else
                    for(int j = 0; j < blockSize; j++)
                        *(dedyData + beg + j) = -(*(goldData + beg + j)) / 
                                                 (*(outputData + beg + j));
            }
        }
    }
    else {
        DTYPE * weightData = (DTYPE*)weight->data;
        if(padding == NULL) {
            for(int i = 0; i < blockNum; i++) {
                int beg = i * blockSize;
                for(int j = 0; j < blockSize; j++)
                    *(dedyData + beg + j) = -(*(weightData + j)) * 
                                             (*(goldData + beg + j)) / 
                                             (*(outputData + beg + j));
            }
        }
        else {
            DTYPE * paddingData = (DTYPE*)padding->data;
            for(int i = 0; i < blockNum; i++) {
                int beg = i * blockSize;
                if(*(paddingData + i) == 0)
                    memset(dedyData + beg, 0, blockSize * unitSize);
                else
                    for(int j = 0; j < blockSize; j++) {
                        *(dedyData + beg + j) = -(*(weightData + j)) * 
                                                 (*(goldData + beg + j)) / 
                                                 (*(outputData + beg + j));
                }
            }
        }        
    }

    if(padding != NULL) {
        XTensor * tmp(padding);
        _IsZero(padding, tmp);
        int nonZeroNum = _ReduceSumAll(tmp);
        _ScaleAndShiftMe(dedy, (DTYPE)1.0/(DTYPE)nonZeroNum);
        delete tmp;
    }
    else {
        _ScaleAndShiftMe(dedy, (DTYPE)1.0/(DTYPE)blockNum);
    }
}

} // namespace nts(NiuTrans.Tensor)