/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2017, Natural Language Processing Lab, Northestern University. 
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: Xu Chen (email: hello_master1954@163.com) 2018-09-17
 */

#ifndef __CROSSENTROPY_CUH__
#define __CROSSENTROPY_CUH__

#include "../XTensor.h"
#include "../XDevice.h"
#include "CrossEntropy.cuh"
#include "CrossEntropy.h"
#include "../core/reduce/ReduceSumAll.h"
#include "../core/math/Unary.h"
#include "../core/math/ScaleAndShift.h"

namespace nts{ // namespace nts(NiuTrans.Tensor)

/* 
compute the cross entropy loss (cuda kernel) 

>> outputData - the data pointer of output tensor
>> goldData - the data pointer of gold tensor
>> lossData - the data pointer of loss tensor
>> weightData - the data pointer of weight tensor
>> paddingData - the data pointer of padding tensor
>> blockNum - the number of data blocks
>> stride - the size of a data block
*/
__global__
void KernelCrossEntropy(DTYPE * outputData, DTYPE * goldData,
                        DTYPE * lossData, DTYPE * weightData, 
                        DTYPE * paddingData, int blockNum, int blockSize)
{
    /* block id */
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    if(i >= blockNum)
        return;

    int beg = i * blockSize;
    DTYPE tmpLoss = 0;

    if(weightData == NULL) {
        if(paddingData == NULL) {
            tmpLoss = 0;
            for(int j = 0; j < blockSize; j++) 
                tmpLoss += -(*(goldData + beg + j)) * 
                            (DTYPE)log(*(outputData + beg + j));
            *(lossData + i) = tmpLoss;
        }
        else {
            if(*(paddingData + i) == 0)
                *(lossData + i) = tmpLoss;
            else{
                for(int j = 0; j < blockSize; j++)
                    tmpLoss += -(*(goldData + beg + j)) * 
                                (DTYPE)log(*(outputData + beg + j));
                *(lossData + i) = tmpLoss;
            }
        }
    }
    else {
        if(paddingData == NULL) {
            for(int j = 0; j < blockSize; j++)
                tmpLoss += -(*(goldData + beg + j)) * 
                            (DTYPE)log(*(outputData + beg + j)) * 
                            (*(weightData + j));
            *(lossData + i) = tmpLoss;
        }
        else {
            if(*(paddingData + i) == 0)
                *(lossData + i) = tmpLoss;
            else{
                tmpLoss = 0;
                for(int j = 0; j < blockSize; j++)
                    tmpLoss += -(*(goldData + beg + j)) * 
                                (DTYPE)log(*(outputData + beg + j)) * 
                                (*(weightData + j));
                *(lossData + i) = tmpLoss;
            }
        }
    }
}

/* 
compute the cross entropy loss (cuda version) 
loss = sum_{i} (-gold_i * log(output_i))
where gold and output are distributions 
        
>> output - model prediction
>> gold - gold standard
>> loss - compute loss
>> weight - a rescaling weight given to each class
>> padding - specify a target value that is ignored and does not contribute to the loss computation
>> leadingDim - the leading dimension for the output
*/
void _CudaCrossEntropyManual(const XTensor * output, const XTensor * gold,
                             XTensor * loss, const XTensor * weight, 
                             const XTensor * padding, int leadingDim)
{
    int order = output->order;
    int n = leadingDim < 0 ? output->order - 1 : leadingDim;
    int leadingDimSize = output->GetDim(n);

    CheckNTErrors(n >= 0 && n < output->order, 
                 "Wrong leadingDim!");
    CheckNTErrors(XTensor::IsSameShaped(output, gold), 
                 "The output tensor and gold tensor must be of the same size!");
    CheckNTErrors(weight == NULL || weight->unitNum == leadingDimSize, 
                 "Wrong weight tensor!");
    CheckNTErrors(padding == NULL || XTensor::IsSameShaped(padding, loss), 
                 "The loss tensor and padding tensor must be same shape!");
    CheckNTErrors(loss->order == output->order - 1, 
                 "Wrong loss dimension!");
    CheckNTErrors(gold->dataType == DEFAULT_DTYPE && output->dataType == DEFAULT_DTYPE, 
                 "TODO!");
    
    int blockNum = 1;
    int blockSize = 1;
    int stride = 1;

    for(int i = n + 1; i < order; i++)
        stride *= output->GetDim(i);
    
    blockSize = stride * leadingDimSize;
    blockNum = output->unitNum / blockSize;

    int cudaGrids[3];
    int cudaBlocks[3];

    //GDevs.GetCudaThread2D(output->devID, blockNum, blockSize, MAX_INT, cudaGrids, cudaBlocks);
    GDevs.GetCudaThread(output->devID, blockNum, cudaGrids, cudaBlocks);
    
    dim3 blocks(cudaGrids[0], cudaGrids[1]);
    dim3 threads(cudaBlocks[0], cudaBlocks[1]);

    int devIDBackup;
    ProtectCudaDev(output->devID, devIDBackup);
    
    DTYPE * outputData = (DTYPE*)output->data;
    DTYPE * goldData = (DTYPE*)gold->data;
    DTYPE * lossData = (DTYPE*)loss->data;

    if(weight == NULL) {
        if(padding == NULL)
            KernelCrossEntropy<<<dim3(cudaGrids[0]), dim3(cudaBlocks[0]) >>>
                                (outputData, goldData, lossData, 
                                 NULL, NULL,
                                 blockNum, blockSize);
        else
            KernelCrossEntropy<<<dim3(cudaGrids[0]), dim3(cudaBlocks[0]) >>>
                                (outputData, goldData, lossData, 
                                 NULL, (DTYPE*)padding->data,
                                 blockNum, blockSize);
    }
    else {
        if(padding == NULL)
            KernelCrossEntropy<<<dim3(cudaGrids[0]), dim3(cudaBlocks[0]) >>>
                                (outputData, goldData, lossData, 
                                 (DTYPE*)weight->data, NULL,
                                 blockNum, blockSize);
        else
            KernelCrossEntropy<<<dim3(cudaGrids[0]), dim3(cudaBlocks[0]) >>>
                                (outputData, goldData, lossData, 
                                 (DTYPE*)weight->data, (DTYPE*)padding->data,
                                 blockNum, blockSize);
    }
    
    BacktoCudaDev(output->devID, devIDBackup);

}

/*
compute the cross entropy loss (scalar version) 

loss = sum_{i} (-gold_i * log(output_i))
where gold and output are distributions 
        
>> output - model prediction
>> gold - gold standard
>> reduceWay - loss compute way, sum or mean
>> weight - a rescaling weight given to each class
>> padding - specify a target value that is ignored and does not contribute to the loss computation
>> leadingDim - the leading dimension for the output
<< return - the cross entropy loss that is a scalar
*/
DTYPE _CudaCrossEntropyManual(const XTensor * output, const XTensor * gold,
                              LOSS_COMPUTE_WAY reduceWay, const XTensor * weight, 
                              const XTensor * padding, int leadingDim)
{
    DTYPE loss = 0;

    int order = output->order;
    int n = leadingDim < 0 ? output->order - 1 : leadingDim;
    int leadingDimSize = output->GetDim(n);

    CheckNTErrors(n >= 0 && n < output->order, 
                 "Wrong leadingDim!");
    CheckNTErrors(XTensor::IsSameShaped(output, gold), 
                 "The output tensor and gold tensor must be of the same size!");
    CheckNTErrors(weight == NULL || weight->unitNum == leadingDimSize, 
                 "Wrong weight tensor!");
    CheckNTErrors(padding == NULL || padding->order == output->order - 1, 
                 "Wrong padding tensor!");
    CheckNTErrors(gold->dataType == DEFAULT_DTYPE && output->dataType == DEFAULT_DTYPE, 
                 "TODO!");
    
    int * dimSize = new int[output->order - 1];
    for (int i = 0; i < order; i++) {
        if(i < n)
            dimSize[i] = output->dimSize[i];
        else if(i > n)
            dimSize[i - 1] = output->dimSize[i];
    }

    XTensor * lossInter = NewTensor(output->order - 1, dimSize, output->dataType, output->denseRatio, output->devID, output->mem);

    _CudaCrossEntropyManual(output, gold, lossInter, weight, padding, leadingDim);

    loss = _ReduceSumAll(lossInter);

    if(reduceWay == REDUCE_MEAN) {
        int totalNum;
        if(padding == NULL) {
            totalNum = lossInter->unitNum;
        }
        else {
            XTensor * zeroIndicator = NewTensorBuf(output, output->devID, output->mem);
            _IsZero(padding, zeroIndicator);
            totalNum = lossInter->unitNum - (int)_ReduceSumAll(zeroIndicator);
            DelTensorBuf(zeroIndicator);
        }

        loss = loss / (DTYPE)totalNum;
    }

    return loss;
}

/* 
backward computation of cross entropy function (kernel version)

>> dedyData - the data pointer of dedy tensor
>> outputData - the data pointer of output tensor
>> goldData - the data pointer of gold tensor
>> weightData - the data pointer of weight tensor
>> paddingData - the data pointer of padding tensor
>> blockNum - the number of data blocks
>> blockSize - the size of a data block
*/
__global__
void KernelCrossEntropyBackward(DTYPE * dedyData, DTYPE * outputData, DTYPE * goldData,
                                DTYPE * weightData, DTYPE * paddingData,
                                int blockNum, int blockSize)
{
    /* block id */
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    if(i >= blockNum)
        return;

    int beg = i * blockSize;

    if(weightData == NULL) {
        if(paddingData == NULL) {
            for(int j = 0; j < blockSize; j++) 
                *(dedyData + beg + j) = -(*(goldData + beg + j)) / 
                                         (*(outputData + beg + j));
        }
        else {
            if(*(paddingData + i) == 0)
                memset(dedyData + beg, 0, blockSize * sizeof(DTYPE));
            else
                for(int j = 0; j < blockSize; j++)
                    *(dedyData + beg + j) = -(*(goldData + beg + j)) / 
                                             (*(outputData + beg + j));
        }
    }
    else {
        if(paddingData == NULL) {
            for(int j = 0; j < blockSize; j++)
                *(dedyData + beg + j) = -(*(weightData + j)) * 
                                         (*(goldData + beg + j)) / 
                                         (*(outputData + beg + j));
        }
        else {
            if(*(paddingData + i) == 0)
                memset(dedyData + beg, 0, blockSize * sizeof(DTYPE));
            else
                for(int j = 0; j < blockSize; j++) {
                    *(dedyData + beg + j) = -(*(weightData + j)) * 
                                             (*(goldData + beg + j)) / 
                                             (*(outputData + beg + j));
            }
        }
    }
}

/* 
backward computation of cross entropy function 

loss = sum_{i} (-t_i * log(y_i))
dE/dy_i = -t_i / y_i
where E is the error(loss) function that measure the errors in y
with respect to gold standard, and y this the model output

>> dedy - dE/dy (for return)
>> output - model prediction
>> gold - gold standard
>> weight - a rescaling weight given to each class
>> padding - specify a target value that is ignored and does not contribute to the loss computation
>> leadingDim - the leading dimension for the output
*/
void _CudaCrossEntropyBackward(XTensor * dedy, const XTensor * output, const XTensor * gold, 
                               const XTensor * weight, XTensor * padding,
                               int leadingDim)
{
    int order = output->order;
    int n = leadingDim < 0 ? output->order - 1 : leadingDim;
    int leadingDimSize = output->GetDim(n);

    CheckNTErrors(n >= 0 && n < output->order, 
                 "Wrong leading dimension!");
    CheckNTErrors(XTensor::IsSameShaped(dedy, output, gold), 
                 "The output tensor and gold tensor must be of the same size!");
    CheckNTErrors(weight == NULL || weight->unitNum == leadingDimSize, 
                 "Wrong weight tensor!");
    CheckNTErrors(padding == NULL || padding->order == output->order - 1, 
                 "Wrong padding tensor!");
    CheckNTErrors(gold->dataType == DEFAULT_DTYPE && output->dataType == DEFAULT_DTYPE, 
                 "TODO!");
    
    int blockNum = 1;
    int blockSize = 1;
    int stride = 1;

    for(int i = n + 1; i < order; i++)
        stride *= output->GetDim(i);
    
    blockSize = stride * leadingDimSize;
    blockNum = output->unitNum / blockSize;

    int cudaGrids[3];
    int cudaBlocks[3];

    GDevs.GetCudaThread(output->devID, blockNum, cudaGrids, cudaBlocks);
    
    dim3 blocks(cudaGrids[0], cudaGrids[1]);
    dim3 threads(cudaBlocks[0], cudaBlocks[1]);

    int devIDBackup;
    ProtectCudaDev(output->devID, devIDBackup);
    
    DTYPE * dedyData = (DTYPE*)dedy->data;
    DTYPE * outputData = (DTYPE*)output->data;
    DTYPE * goldData = (DTYPE*)gold->data;

    if(weight == NULL) {
        if(padding == NULL)
            KernelCrossEntropyBackward<<<dim3(cudaGrids[0]), dim3(cudaBlocks[0]) >>>
                                        (dedyData, outputData, goldData,
                                         NULL, NULL,
                                         blockNum, blockSize);
        else
            KernelCrossEntropyBackward<<<dim3(cudaGrids[0]), dim3(cudaBlocks[0]) >>>
                                        (dedyData, outputData, goldData,
                                         NULL, (DTYPE*)padding->data,
                                         blockNum, blockSize);
    }
    else {
        if(padding == NULL)
            KernelCrossEntropyBackward<<<dim3(cudaGrids[0]), dim3(cudaBlocks[0]) >>>
                                        (dedyData, outputData, goldData,
                                        (DTYPE*)weight->data, NULL,
                                         blockNum, blockSize);
        else
            KernelCrossEntropyBackward<<<dim3(cudaGrids[0]), dim3(cudaBlocks[0]) >>>
                                        (dedyData, outputData, goldData,
                                        (DTYPE*)weight->data, (DTYPE*)padding->data,
                                         blockNum, blockSize);
    }

    if(padding != NULL) {
        XTensor * tmp(padding);
        _IsZero(padding, tmp);
        int nonZeroNum = _ReduceSumAll(tmp);
        _ScaleAndShiftMe(dedy, (DTYPE)1.0/(DTYPE)nonZeroNum);
        delete tmp;
    }
    else {
        _ScaleAndShiftMe(dedy, (DTYPE)1.0/(DTYPE)blockNum);
    }
    
    BacktoCudaDev(output->devID, devIDBackup);

}

} // namespace nts(NiuTrans.Tensor)

#endif // __CROSSENTROPY_CUH__